/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*      ________________________________________________________________        */
/*      | H Y T E C   8 4 1 7  2 4 B i t   A D C  A s y n  D r i v e r |        */
/*           -----------------------------------------------------------        */
/*                                                                              */
/*  Source file name      :-      drvHy8415Asyn.cpp                             */
/*                                                                              */
/*  Initial creation date :-      16-Nov-2011                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8415 DAC EPICS support. It is based   */
/*  on Hy8417ipasyn driver code. It is osi complaint.		                    */
/*                                                                              */
/*  (C)2011 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    16/11/11    |  JC		  | 1.0     | Initial creation with new asyn model  */    
/*    24/05/12    |  JC		  | 1.1     | Added clearing rebooting flag at the	*/    
/*                |           |         | beginning to avoid self hanging for   */    
/*    26/11/12    |  JC		  | 1.2     | Fixed minus typos						*/    
/*    10/01/14    |  JC		  | 1.3     | Added function generator. Fixed		*/    
/*                |           |         | parameter index 15 warning msg        */
/*    10/09/14    |  JC		  | 1.4     | Fixed	clock rate report error			*/    
/*    06/10/14    |  JC		  | 1.5     | Modified to support 64bit Linux based	*/    
/*                |  M.Rivers |         | on Mark Rivers' driverTable model     */    
/*                |           |         | Note: need V2.6 of drvHyLinuxDriver.c */    
/*                |           |         | in ipac module to work with 64bit     */    
/*                |           |         | Linux		                            */    
/*                |           |         |    		                            */
/********************************************************************************/

#define DRIVER_VERSION	105

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8415Asyn.h"

static Hy8415Asyn *driverTable[MAX_CARDS];
static int numCards;

/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 *
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *         mode = 0: register, 1: triggered
 *      range   = 0: -10V ~ +10V, 1:-5V ~ +5V, 2: 0 ~ 10V, 3: 0 ~ 5V 
 *    clockRate = 0: 1Hz,    1: 2Hz,    2: 5Hz,    3: 10Hz,
 *				  4: 20Hz,   5: 50Hz,   6: 100Hz,  7: 200Hz,
 *				  8: 500Hz,  9: 1kHz,  10: 2kHz,  11: 5kHz,
 *				  12: 10khz, 13: 20Khz, 14: 50kHz
 *     extClock = 0: internal, 1: extrenal,
 */
    
Hy8415Asyn::Hy8415Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode, 
		int range, int clockRate, int extClock)    
		: asynPortDriver(portName, MAX_DAC_CHANNELS, NUM_PARAMS, 
		asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
  	char lstr[10];
  	USHORT idmem;
    int ishytec;
    int ismodel;
	int ver;

	/* normal data channels definition */
    createParam(hy8415DataString, 			asynParamInt32, 	&hy8415Data_);
    createParam(hy8415FloatDataString, 		asynParamFloat64, 	&hy8415FloatData_);

	/* controls definition */
    createParam(hy8415SetModeString, 		asynParamInt32, 	&hy8415SetMode_);
    createParam(hy8415NoOfUpdateString,		asynParamInt32, 	&hy8415NoOfUpdate_);
    createParam(hy8415ContinuousString,		asynParamInt32, 	&hy8415Continuous_);
    createParam(hy8415SoftTriggerString, 	asynParamInt32, 	&hy8415SoftTrigger_);
    createParam(hy8415SetClockRateString, 	asynParamInt32, 	&hy8415SetClockRate_);
    createParam(hy8415SetExtClockString, 	asynParamInt32, 	&hy8415SetExtClock_);
    createParam(hy8415RangeString, 			asynParamInt32, 	&hy8415Range_);
    createParam(hy8415SetFormatString, 		asynParamInt32, 	&hy8415SetFormat_);
    createParam(hy8415SetMultiTriggerString,asynParamInt32, 	&hy8415SetMultiTrigger_);
    createParam(hy8415RepeatPatternString, 	asynParamInt32, 	&hy8415RepeatPattern_);
    createParam(hy8415Set16bitString, 		asynParamInt32, 	&hy8415Set16bit_);
    createParam(hy8415ZeroDACString, 		asynParamInt32, 	&hy8415ZeroDAC_);


    createParam(hy8415CounterString,		asynParamInt32, 	&hy8415Counter_);
    createParam(hy8415TriggerStatusString,	asynParamInt32, 	&hy8415TriggerStatus_);
    createParam(hy8415FWVersionString, 		asynParamInt32, 	&hy8415FWVersion_);
    createParam(hy8415DriverVersionString,	asynParamInt32, 	&hy8415DriverVersion_);
    createParam(hy8415SupportString, 		asynParamInt32, 	&hy8415Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		 = carrierNum;
    this->ipSlotNum			 = ipSlotNum;
    this->vectorNum          = vectorNum;
    this->range          	 = range;
    this->mode          	 = mode;
    this->clockrate        	 = clockRate;
    this->extclock           = extClock;
	setIntegerParam(hy8415SetMode_, mode);
	setIntegerParam(hy8415SetClockRate_, clockRate);
	setIntegerParam(hy8415SetExtClock_, extClock);
	setIntegerParam(hy8415TriggerStatus_, WAITING_TRIGGER);

	//initial (default) settings
    this->noofupdate       	 = 0;
    this->continuous       	 = 0;
    this->format	       	 = 0;
    this->multitrigger     	 = 0;
    this->repeatpattern    	 = 0;
    this->bitFormat       	 = 0;

	/* Map Register address spaces  */
	this->dac = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->dac == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8415: Uhy8415NoOfUpdatenable to register IO space\n");
        return;
    }

    /* map the memory address spaces  */
    this->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (this->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHy8415: Unable to register memory space\n");
        return;
    }

    /* Compare the HY8415 prom header, manufacturer's ID and card model
     * number with the expected values. */
  	idmem =  *((USHORT*)this->dac->prom.vita4);
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->dac->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->dac->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->dac->prom.vita4)) == 0;
    ishytec = this->dac->prom.hytecIdHigh == HYTECID_HI  &&
                  this->dac->prom.hytecIdLow == HYTECID_LO;
    ismodel = this->dac->prom.model == PROM_MODEL;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %#08x\n", (int)&this->dac->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->dac->prom.hytecIdHigh, (int)this->dac->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->dac->prom.model, (int)this->dac->prom.revision);

	ver = (int)this->dac->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;
	setIntegerParam(hy8415FWVersion_, ver);

    printf("PROM: Serial number      - %hu\n", (int)this->dac->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->dac->prom.driverIdHigh, (int)this->dac->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->dac->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->dac->prom.byteCount, (int)this->dac->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}
   
    /* set driver version */
	setIntegerParam(hy8415DriverVersion_, DRIVER_VERSION);
	
    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    /* only set up interrupt, work queue and mutex stuff when it is trigger mode */
    if(this->mode)
    {
    	this->dac->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
    	driverTable[numCards] = this;
    	numCards++;
		if (ipmIntConnect(this->carrierNum, this->ipSlotNum, this->vectorNum, INTSRC, numCards-1) < 0) {
			printf("configHy8415: failure connecting to interrupt\n");
      	   return;
    	}

    	ipmIrqCmd(this->carrierNum, this->ipSlotNum, 0, ipac_irqEnable);

    	this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_DAC_CHANNELS)*sizeof(int));
    	
    	if (epicsThreadCreate("Hy8415intQueuedTask",
                           epicsThreadPriorityLow,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intQueuedTaskC,
                           (void *)this) == NULL)
		{	
      		printf("configHy8415: failure creating message queue.\n");
      		return;
		}
    }
    
	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/

	this->rebooting = 0;

	setupDAC();

	// tell listeners the changes
	callParamCallbacks();
   	epicsAtExit(rebootCallbackC, this);
}

// Extension Init
int Hy8415Asyn::Hy8415ExtInit(int continuous, int format, int multitrigger, 
		int repeatpattern, int set16bit)
{
	if((continuous < 0 || continuous > 1) ||
			(format < 0 || format > 1) ||
			(multitrigger < 0 || multitrigger > 1) ||
			(repeatpattern < 0 || repeatpattern > 1) ||
			(set16bit < 0 || set16bit > 1)) return asynError;
	
    this->noofupdate       	 = 0;										//this will be set by waveform data download.
    this->continuous       	 = continuous;
    this->format	       	 = format;
    this->multitrigger     	 = multitrigger;
    this->repeatpattern    	 = repeatpattern;
    this->bitFormat    	 	 = set16bit;
    
	setNumberUpdate( noofupdate );
	setContinuousUpdate( continuous );
    setFormat( format );
	setMultiTrigger( multitrigger );
	setRepeat( repeatpattern );
	set16Bit( set16bit );

	return(asynSuccess);
}

int Hy8415Asyn::Hy8415Function(int type, double volt, int samples, int mask)
{
	int i, j, halfmem, bitformat;
	double x,z;
	ULONG y;
	
	getIntegerParam(hy8415Set16bit_, &bitformat);

	if(((bitformat == 1) && (samples > CHANNELBANKSIZE*2)) ||
		((bitformat == 0) && (samples > CHANNELBANKSIZE))) return asynError;
	
	this->noofupdate = samples;
	setIntegerParam(hy8415NoOfUpdate_, this->noofupdate);
	setNumberUpdate( samples );
	
	switch(type)
	{
	case 0:
		for(i=0; i<samples; i++)
		{
			if(bitformat == 0)				//18bit
			{
				z = 0x3FFFF*volt*2/20;
				x = (z*sin(i*2*PI/samples) + z)/2;
				y = (ULONG)x;
				if(i > CHANNELBANKSIZE/2) 
					halfmem = BANKSIZE/2;
				else
					halfmem = 0;
				for(j=0; j<16; j++)
				{
					if((1 << j) & mask)
					{
						this->convMemoryBase[i*2+j*CHANNELBANKSIZE+halfmem] = (USHORT) (y & 0xFFFF);
						this->convMemoryBase[i*2+1+j*CHANNELBANKSIZE+halfmem] = (USHORT) (y >> 16);
					}
					else
					{
						this->convMemoryBase[i*2+j*CHANNELBANKSIZE+halfmem] = 0xFFFF;
						this->convMemoryBase[i*2+1+j*CHANNELBANKSIZE+halfmem] = 0x01;
					}
				}
			}else
			{
				z = 0xFFFF*volt*2/20;
				x = (z*sin(i*2*PI/samples) + z)/2;
				y = (ULONG)x;
				if(i > CHANNELBANKSIZE) 
					halfmem = BANKSIZE/2;
				else
					halfmem = 0;
				for(j=0; j<16; j++)
				{
					if((1 << j) & mask)
						this->convMemoryBase[i+j*CHANNELBANKSIZE+halfmem] = (USHORT) (y & 0xFFFF);
					else
						this->convMemoryBase[i+j*CHANNELBANKSIZE+halfmem] = 0x8000;
				}
			}
		}
		break;
	case 1:
		for(i=0; i<samples; i++)
		{
			if(bitformat == 0)				//18bit
			{
				if(i<samples/2)
				{
					x = (0x3FFFF*volt*2/20);
					y = (USHORT) x;
				}else
				{
					y = 0;
				}
				if(i > CHANNELBANKSIZE/2) 
					halfmem = BANKSIZE/2;
				else
					halfmem = 0;
				for(j=0; j<16; j++)
				{
					if((1 << j) & mask)
					{
						this->convMemoryBase[i*2+j*CHANNELBANKSIZE+halfmem] = (USHORT) (y & 0xFFFF);
						this->convMemoryBase[i*2+1+j*CHANNELBANKSIZE+halfmem] = y >> 16;
					}
					else
					{
						this->convMemoryBase[i*2+j*CHANNELBANKSIZE+halfmem] = 0xFFFF;
						this->convMemoryBase[i*2+1+j*CHANNELBANKSIZE+halfmem] = 0x01;
					}
				}
			}else
			{
				if(i<samples/2)
				{
					x = (0xFFFF*volt*2/20);
					y = (USHORT) x;
				}else
				{
					y = 0;
				}
				if(i > CHANNELBANKSIZE) 
					halfmem = BANKSIZE/2;
				else
					halfmem = 0;
				for(j=0; j<16; j++)
				{
					if((1 << j) & mask)
						this->convMemoryBase[i+j*CHANNELBANKSIZE+halfmem] = (USHORT) (y & 0xFFFF);
					else
						this->convMemoryBase[i+j*CHANNELBANKSIZE+halfmem] = 0x8000;
				}
			}
		}
		break;
	default:
		return asynError;
	}
	return asynSuccess;
}

/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8415Asyn::report(FILE *fp, int details)
{
	int mode, range, format, extclock, clockrate, fwver, swver, noofupdate, cont, multi, repeat, set16;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p\n", 
            this->portName, this->carrierNum, this->ipSlotNum, (void *)this->dac);
    if (details == 1) 
    {
		getIntegerParam(hy8415SetMode_, &mode);
		getIntegerParam(hy8415Range_, &range);
		getIntegerParam(hy8415SetClockRate_, &clockrate);
		getIntegerParam(hy8415SetExtClock_, &extclock);
		getIntegerParam(hy8415SetFormat_, &format);
		getIntegerParam(hy8415NoOfUpdate_, &noofupdate);
		getIntegerParam(hy8415Continuous_, &cont);
		getIntegerParam(hy8415SetMultiTrigger_, &multi);
		getIntegerParam(hy8415RepeatPattern_, &repeat);
		getIntegerParam(hy8415Set16bit_, &set16);
		getIntegerParam(hy8415FWVersion_, &fwver);
		getIntegerParam(hy8415DriverVersion_, &swver);
		
		fprintf(fp, "  RegBase=%x MemBase=%x\n", (int)this->dac, (int)this->convMemoryBase);
        fprintf(fp, "  Current CSR=%x\n", this->dac->reg.csr);
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum);
        fprintf(fp, "  Mode=%s\n", mode==0?"Register":"Trigger");
        fprintf(fp, "  Range=%s\n", range==0?"+/-10V":(range==1?"+/-5V":(range==2?"0-10V":"0-5V")));
        fprintf(fp, "  Clock Source=%s\n", extclock==0?"Internal":"External");
        fprintf(fp, "  Clock Rate=%d\n", clockrate);
        fprintf(fp, "  Format=%s\n", format==1?"2's complement":"Straight binary");
        fprintf(fp, "  DAC data bit=%s\n", set16==0?"18bit":"16bit");       	
        fprintf(fp, "  Firmware Version=%d\n", fwver);
        fprintf(fp, "  Driver Version=%d\n", swver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);
        
        if(mode == 1)
        {
            fprintf(fp, "  Number of update=%d\n", noofupdate);
            fprintf(fp, "  Continuous update=%s\n", cont==0?"No":"Yes");
            fprintf(fp, "  Multi Trigger=%s\n", multi==0?"No":"Yes");       	
            fprintf(fp, "  Multi Repeat Pattern=%s\n", repeat==0?"No":"Yes");       	
        }
    }
    asynPortDriver::report(fp, details);
}

/*****         asynInt32 methods       ******/
asynStatus Hy8415Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	USHORT csr = this->dac->reg.csr, high = 0, low = 0;
	int timeout = 10000, bitformat;
	
    this->getAddress(pasynUser, &channel);
	getIntegerParam(hy8415Set16bit_, &bitformat);
	
	/* decide what to do depending on which command we received */
	if (function == hy8415Data_)
	{
		if(!bitformat)													/* 18bit DAC */
		{
			low = value & 0xFFFF;
			high = value >> 16;
			this->dac->reg.data[channel*2] = low;
			this->dac->reg.data[channel*2 + 1] = high;
//			this->dac->reg.data[4] = low;
//			this->dac->reg.data[5] = high;
		}else															/* 16bit DAC */
		{
			this->dac->reg.data[channel] = (USHORT)value;			
		}
		
		while((csr & CSR_BS) && (timeout)){
			csr = this->dac->reg.csr;
			timeout--;
		}
		if(timeout) this->dac->reg.csr |= CSR_ARM;
		else status = asynError;
//testing		
/*printf("chl=%d csr=%x value=%x high=%x low=%x timeout=%d\n", channel, this->dac->reg.csr, 
				value, this->dac->reg.data[channel*2 + 1], this->dac->reg.data[channel*2], timeout);*/
	}
	else if (function == hy8415SetMode_)
		status = setMode( value );
	else if (function == hy8415NoOfUpdate_)
		status = setNumberUpdate( value );
	else if (function == hy8415Continuous_)
		status = setContinuousUpdate( value );
	else if (function == hy8415SoftTrigger_)
		status = softwareTrigger();
	else if (function == hy8415SetClockRate_)
		status = setClockRate( value );
	else if (function == hy8415SetExtClock_)
		status = setExtClock( value );
	else if (function == hy8415Range_)
		status = setRange( value );
	else if (function == hy8415SetFormat_)
		status = setFormat( value );
	else if (function == hy8415SetMultiTrigger_)
		status = setMultiTrigger( value );
	else if (function == hy8415RepeatPattern_)
		status = setRepeat( value );
	else if (function == hy8415Set16bit_)
		status = set16Bit( value );
	else if (function == hy8415ZeroDAC_)
		zeroDAC();

	return( status );
}

asynStatus Hy8415Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int triggerState, ver, mode;

	if (this->rebooting) epicsThreadSuspendSelf();
	getIntegerParam(hy8415TriggerStatus_, &triggerState);
	getIntegerParam(hy8415SetMode_, &mode);

	this->getAddress(pasynUser, &channel);
	
	if(function == hy8415Support_)
		*value = DRIVER_SUPPORT;

	else if(function == hy8415FWVersion_)
	{
		getIntegerParam(hy8415FWVersion_, &ver);
        *value = ver;                                               	 		/* firmware version */
	}
	else if(function == hy8415DriverVersion_)
	{
		getIntegerParam(hy8415DriverVersion_, &ver);
        *value = ver;                                               	 		/* driver version */
	}
	else if(function == hy8415TriggerStatus_)
	{
        if (mode == hy8415Trigger) 
        	*value = triggerState;
	}
	else if(function == hy8415Counter_)
		*value = (epicsInt32) this->dac->reg.pointer;
	else
		asynPortDriver::readInt32(pasynUser, value);
	
	return( status );
}

asynStatus Hy8415Asyn::writeInt32Array(asynUser *pasynUser, epicsInt32 *value, size_t nElements)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int i, bitformat, v=0;
	USHORT low, high;

	if (this->rebooting) epicsThreadSuspendSelf();
	getIntegerParam(hy8415Set16bit_, &bitformat);

	this->getAddress(pasynUser, &channel);
	
	if (function == hy8415Data_)
	{
		for(i=0; i<(int)nElements; i++)
		{
			if(bitformat)
			{
				if(i < CHANNELBANKSIZE)
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i] = *(value+i);
				else
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i] = *(value+i);					
			}else
			{
				v = value[i];
				low = (USHORT) (v & 0xFFFF);
				high =(USHORT) (v >> 16);
				if(i < (CHANNELBANKSIZE / 2))
				{
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2] = low;
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2 + 1] = high;
				}else
				{
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2] = low;
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2 + 1] = high;					
				}
			}
		}
		setNumberUpdate( (int)nElements );
	}
	return( status );
}

/* asynInt32Array methods */
asynStatus Hy8415Asyn::readInt32Array(asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int i, bitformat;
	USHORT low, high;

	if (this->rebooting) epicsThreadSuspendSelf();
	getIntegerParam(hy8415Set16bit_, &bitformat);

	this->getAddress(pasynUser, &channel);

	if (function == hy8415Data_)
	{
		for(i=0; i<(int)nelements; i++)
		{
			if(bitformat)
			{
				if(i < CHANNELBANKSIZE)
					low = (USHORT)((this->convMemoryBase + channel * CHANNELBANKSIZE)[i]);
				else
					low = (USHORT)((this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i]);
				*(data+i) = low;
			}else
			{
				if(i < (CHANNELBANKSIZE / 2))
				{
					low = (USHORT)((this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2]);
					high = (USHORT)((this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2 + 1]);
				}else
				{
					low = (USHORT)((this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2]);
					high = (USHORT)((this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2 + 1]);					
				}
				*(data+i) = (high << 16) + low;
			}
		}
	}
	return( status );
}

asynStatus Hy8415Asyn::getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 65535;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}


asynStatus Hy8415Asyn::writeFloat64(asynUser *pasynUser, epicsFloat64 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int timeout = 10000, bitformat = 0;
	USHORT csr = this->dac->reg.csr, high = 0, low = 0;
	
    if (this->rebooting) epicsThreadSuspendSelf();

	this->getAddress(pasynUser, &channel);
	getIntegerParam(hy8415Set16bit_, &bitformat);
	
	/* decide what to do depending on which command we received */
	if (function == hy8415FloatData_)
	{
		if(!bitformat)													/* 18bit DAC */
		{
			low = fromVoltage(value) & 0xFFFF;
			high = fromVoltage(value) >> 16;
			this->dac->reg.data[channel*2] = low;
			this->dac->reg.data[channel*2 + 1] = high;
		}else															/* 16bit DAC */
		{
			this->dac->reg.data[channel] = (USHORT)fromVoltage(value);			
		}
		while((csr & CSR_BS) && (timeout)){
			csr = this->dac->reg.csr;
			timeout--;
		}
		if(timeout) this->dac->reg.csr |= CSR_ARM;
		else status = asynError;
	}
//testing		
/*
if(!bitformat)
	printf("chl=%d csr=%x value=%f high=%x low=%x addh=%x addl=%x timeout=%d\n", channel, this->dac->reg.csr, 
		value, this->dac->reg.data[channel*2 + 1], this->dac->reg.data[channel*2], 
		&this->dac->reg.data[channel*2 + 1], &this->dac->reg.data[channel*2],timeout);
else
	printf("chl=%d csr=%x value=%f regvalue=%x timeout=%d\n", channel, this->dac->reg.csr, 
		value, this->dac->reg.data[channel], timeout);
*/
	return( status );
}

asynStatus Hy8415Asyn::writeFloat64Array(asynUser *pasynUser, epicsFloat64 *value, size_t nElements)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
	int i, bitformat;
	USHORT low, high;

	if (this->rebooting) epicsThreadSuspendSelf();
	getIntegerParam(hy8415Set16bit_, &bitformat);
	
	this->getAddress(pasynUser, &channel);
	
	if (function == hy8415FloatData_)
	{
		for(i=0; i<(int)nElements; i++)
		{
			if(bitformat)
			{
				if(i < CHANNELBANKSIZE)
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i] = (USHORT) fromVoltage(*(value+i));
				else
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i] = (USHORT) fromVoltage(*(value+i));					
			}else
			{
				low = (fromVoltage(*(value + i))) & 0xFFFF;
				high = (fromVoltage(*(value + i))) >> 16;
				if(i < CHANNELBANKSIZE / 2)
				{
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2] = low;
					(this->convMemoryBase + channel * CHANNELBANKSIZE)[i*2 + 1] = high;
				}else
				{
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2] = low;
					(this->convMemoryBase + BANKSIZE/2 + channel * CHANNELBANKSIZE)[i*2 + 1] = high;					
				}
/*printf("low=%x high=%x\n", low, high);*/
			}
		}
		setNumberUpdate( (int)nElements );
	}
	return( status );
}


/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8415Asyn::setupDAC()
{
	USHORT csr = 0;
	if (this->rebooting)  epicsThreadSuspendSelf();
	this->dac->reg.csr = 0; 											/* clear csr */   		

	if(this->mode) csr = CSR_EE;
	this->dac->reg.csr = csr; 													/* initial csr */   
	this->dac->reg.pointer = 0; 												/* clear pointer */   
	setClockRate( this->clockrate );	          
	setExtClock( this->extclock );	              
	setFormat( this->format );
	setRange( this->range );
	set16Bit(this->bitFormat);
	if(this->mode)
	{
		setNumberUpdate(this->noofupdate);
		setContinuousUpdate(this->continuous);
		setMultiTrigger(this->multitrigger);
		setRepeat(this->repeatpattern);
	}
	setMode( this->mode );														/* after this, DAC is armed */
	zeroDAC();																	/* 0 all outputs */
}

asynStatus Hy8415Asyn::setClockRate( int rate  )
{
	int mode, extClock;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);
	getIntegerParam(hy8415SetExtClock_, &extClock);

	/* for internal clock only */
	if((mode == hy8415Register) || (extClock == 1)) return(asynError);
	
	/* for internal clock */
    if ((rate < 0) || (rate > 14)) return( asynError );

    this->dac->reg.clockRate = rate;
	setIntegerParam(hy8415SetClockRate_, rate);
	if(mode) resetTriggerState();
	callParamCallbacks();
	return(asynSuccess);
}

asynStatus Hy8415Asyn::setExtClock( int ext )
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);

    if ((ext < 0) || (ext > 1)) return( asynError );

	if(mode)
	{
		if( ext )                                                   	/* use external clock */
			this->dac->reg.csr |= CSR_XC;
		else
			this->dac->reg.csr &= ~CSR_XC;
		if(mode) resetTriggerState();
	}else
		return asynError;
	
	setIntegerParam(hy8415SetExtClock_, ext);
	callParamCallbacks();
    return( asynSuccess );
}

asynStatus Hy8415Asyn::softwareTrigger()
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);
 
    if(mode != hy8415Trigger ) 
        return asynError;                                           /* software trigger at realtime only applies to trigger mode */
    else                                           					/* only trigger when enabled */
    {
        this->dac->reg.csr &= ~CSR_ST;                           	/* clear software trigger latch */
        this->dac->reg.csr |= CSR_ST;                            	/* software trigger to kick off the acquisition */
        this->dac->reg.csr &= ~CSR_ST;                           	/* clear software latch to enable hardware trigger afterwards */
		setIntegerParam(hy8415TriggerStatus_, TRIGGERRED_UPDATING);
  		callParamCallbacks();
    }
    return( asynSuccess );
}

asynStatus Hy8415Asyn::setMode( int mode )
{
	if (this->rebooting)  epicsThreadSuspendSelf();
    
    if ((mode < 0) || (mode > 1)) return( asynError );
	
    if (mode == hy8415Register)
        this->dac->reg.csr &= ~CSR_EX;                    			/* disable trigger */
    else
    {
    	resetTriggerState();
        this->dac->reg.csr |= CSR_EX;            			/* enable trigger */
    }
    
	setIntegerParam(hy8415SetMode_, mode);
	callParamCallbacks();
    return( asynSuccess );
}

asynStatus Hy8415Asyn::setNumberUpdate( int number )
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

    if ((number < 0) || (mode > CHANNELBANKSIZE)) return( asynError );

	getIntegerParam(hy8415SetMode_, &mode);
 
	if (mode)                                                     	/* if trigger mode */
	{
		this->dac->reg.updateNo = number;                    		/* set number of updates */
        setIntegerParam(hy8415NoOfUpdate_, number);
    	if(mode) resetTriggerState();
    }
	else                                                            /* if not trigger mode */
        return asynError;                                           /* set error */
	
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::setContinuousUpdate(int cont)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);

	if ((cont < 0) || (cont > 1)) return( asynError );
 
	if (mode)                                                     	/* if trigger mode */
	{
		if(cont)
			this->dac->reg.csr |= CSR_CONT;                    		/* set continuous function generation */
		else
			this->dac->reg.csr &= ~CSR_CONT;                    	/* stop continuous function generation */
		if(mode) resetTriggerState();
        setIntegerParam(hy8415Continuous_, cont);
    }
	else                                                            /* if not trigger mode */
        return asynError;                                           /* set error */
	
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::setRange(int range)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);

	if ((range < 0) || (range > 3)) return( asynError );

    this->dac->reg.extcsr &= 0xFFFC;	                				/* clear range to defult +/-10V */
	switch(range)
	{
	case 0:
        this->dac->reg.extcsr |= ACR_PM10V;	                			/* set +/-5V range */
		break;
	case 1:
        this->dac->reg.extcsr |= ACR_PM5V;	                			/* set +/-5V range */
		break;
	case 2:
        this->dac->reg.extcsr |= ACR_ZT10V;	                			/* set +/-5V range */
		break;
	case 3:
        this->dac->reg.extcsr |= ACR_ZT5V;	                			/* set +/-5V range */
		break;
	}
		
	if(mode) resetTriggerState();

	setIntegerParam(hy8415Range_, range);
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::setFormat(int format)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);

	if ((format < 0) || (format > 1)) return( asynError );

	if(format)
        this->dac->reg.extcsr |= ACR_BTC;	                			/* 2's complement */
	else
        this->dac->reg.extcsr &= ~ACR_BTC;	                			/* straight binary */

	if(mode) resetTriggerState();

	setIntegerParam(hy8415SetFormat_, format);
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::setMultiTrigger(int multi)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);
 
	if ((multi < 0) || (multi > 1)) return( asynError );

	if (mode)                                                     	/* if multi trigger mode */
	{
		if(multi)
			this->dac->reg.extcsr |= ACR_MT;                    	/* set multi trigger */
		else
			this->dac->reg.extcsr &= ~ACR_MT;                    	/* stop multi trigger */
	    if(mode) resetTriggerState();
        setIntegerParam(hy8415SetMultiTrigger_, multi);
    }
	else                                                            /* if not trigger mode */
        return asynError;                                           /* eturn error */
	
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::setRepeat(int repeat)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);
 
	if ((repeat < 0) || (repeat > 1)) return( asynError );

	if (mode)                                                     	/* if repeat mode */
	{
		if(repeat)
			this->dac->reg.extcsr |= ACR_RP;                    	/* set repeat pattern */
		else
			this->dac->reg.extcsr &= ~ACR_RP;                    	/* stop repeat pattern */
	    if(mode) resetTriggerState();
        setIntegerParam(hy8415RepeatPattern_, repeat);
    }
	else                                                            /* if not trigger mode */
        return asynError;                                           /* return error */
	
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::set16Bit(int bit)
{
	int mode;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8415SetMode_, &mode);

	if ((bit < 0) || (bit > 1)) return( asynError );

	if(bit)
		this->dac->reg.extcsr |= ACR_EN16;                    		/* set 16bit DAC */
	else
		this->dac->reg.extcsr &= ~ACR_EN16;                    		/* restore 18bit DAC */
    setIntegerParam(hy8415Set16bit_, bit);

    if(mode) resetTriggerState();
	callParamCallbacks();
	return( asynSuccess );
}

asynStatus Hy8415Asyn::zeroDAC()
{
    this->dac->reg.csr |= CSR_SSU;                  					/* Zero DAC outputs */		
    this->dac->reg.csr &= ~CSR_SSU;                  					/* clear flag */		
	return( asynSuccess );
}

asynStatus Hy8415Asyn::resetTriggerState()
{
    this->dac->reg.csr &= ~CSR_ARM;                  					/* Dis-arm */		
    this->dac->reg.pointer = 0;                  						/* clear pointer */		
    this->dac->reg.csr |= CSR_ARM;                  					/* Re-arm */		
    setIntegerParam(hy8415TriggerStatus_, WAITING_TRIGGER);
	callParamCallbacks();
	return( asynSuccess );
}

void Hy8415Asyn::rebootCallback()
{
	zeroDAC();															/* 0 all outputs */
	this->dac->reg.csr = 0; 											/* clear csr */   		
	this->dac->reg.pointer = 0; 										/* clear pointer */   		
    this->rebooting = 1;				                            	/* flag we are rebooting */
}

/*********************************************************/
/*                                                      */
/* Floating Voltage Convertion Routines                */
/*                                                    */
/*****************************************************/

/* this routine convert the raw dac value to voltage, -10V ~ +10V or -5V ~ +5V as per the range setting.  */
int Hy8415Asyn::fromVoltage( double data )
{
	int range, format, bit, v = 0;
	
	getIntegerParam(hy8415Range_, &range);
	getIntegerParam(hy8415SetFormat_, &format);
	getIntegerParam(hy8415Set16bit_, &bit);

	if(bit == 0)															//18bit DAC
	{
		switch(range)														// decide range
		{
		case 0:																//+/-10V
			data = (data < -10) ? -10 : ((data > 10) ? 10 : data);
			v = (int)(data * 0x1FFE0 / 10 + 0x20000);
			break;
		case 1:																//+/-5V
			data = (data < -5) ? -5 : ((data > 5) ? 5 : data);
			v = (int)(data * 0x1FFC0 / 5 + 0x20000);
			break;
		case 2:																//0-10V
			data = (data < 0) ? 0 : ((data > 10) ? 10 : data);
			v = (int)(data * 0x3FFC0 / 10);
			break;
		case 3:																//0-5V
			data = (data < 0) ? 0 : ((data > 5) ? 5 : data);
			v = (int)(data * 0x3FFC0 / 5);
			break;
		}
		if(format)															//2's complement
		{
			if(v > 0x20000)
				v = v - 0x20000;
			else
				v = ~(0x20000 - v) - 1;
		}
	}
	else																	//16bit DAC
	{
		switch(range)														// decide range
		{
		case 0:																//+/-10V
			data = (data < -10) ? -10 : ((data > 10) ? 10 : data);
			v = (int)(data * 0x7FF8 / 10 + 0x8000);
			break;
		case 1:																//+/-5V
			data = (data < -5) ? -5 : ((data > 5) ? 5 : data);
			data = (int)(data * 0x7FF0 / 5 + 0x8000);
			break;
		case 2:																//0-10V
			data = (data < 0) ? 0 : ((data > 10) ? 10 : data);
			v = (int)(data * 0xFFF0 / 10);
			break;
		case 3:																//0-5V
			data = (data < 0) ? 0 : ((data > 5) ? 5 : data);
			v = (int)(data * 0xFFF0 / 5);
			break;
		}
		if(format)														//2's complement
		{
			if(v > 0x8000)
				v = v - 0x8000;
			else
				v = ~(0x8000 - v) - 1;
		}
	}
	return v;
}



/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8415Asyn *pPvt = (Hy8415Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int cardNumber)
{
    Hy8415Asyn *pPvt = driverTable[cardNumber];
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8415Asyn *pPvt = (Hy8415Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8415Asyn::intSR()
{
    int data[2];

    this->dac->reg.csr &= ~CSR_EE;                  								/* disable interrupt */

	if (this->dac->reg.csr & CSR_CC)     											/* Conversion complete */
    {
        /* Wake up queue task which calls callback routines */
         if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
         	messagesSent++;
         else
         	messagesFailed++;
    }
//    callParamCallbacks();
}

void Hy8415Asyn::intQueuedTask()
{
    int data[2];

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
		setIntegerParam(hy8415TriggerStatus_, UPDATE_COMPLETE);
	    this->dac->reg.pointer = 0;                  								/* clear pointer */
	    this->dac->reg.csr &= ~CSR_CC;                  							/* clear convert complete */
	    this->dac->reg.csr |= CSR_EE;                  								/* enable interrupt */
        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);
    }
    callParamCallbacks();
}

